#!/usr/bin/env python
# -*- coding: utf-8 -*-

# # 2A.i - Parallélisation locale (énoncé)
# 
# Parallélisation avec [joblib](https://pythonhosted.org/joblib/).

# In[1]:


from jyquickhelper import add_notebook_menu
add_notebook_menu()


# La parallélisation des calculs permet d'accélèrer l'exécution d'un programme au prix de deux contraintes :
# 
# * La [synchronisation](https://en.wikipedia.org/wiki/Synchronization_%28computer_science%29) : écrire un programme avec des calculs en parallèle est plus complexe. Il faut souvent attendre que des processus s'exécutant en parallèle se terminent avant de passer à l'étape suivante avec l'ensemble des résultats.
# * La [communication](https://en.wikipedia.org/wiki/Inter-process_communication) : il faut communiquer aux différents processus d'exécutant en parallèle les tâches qu'ils doivent exécuter. Selon la parallélisation choisie, le coût n'est pas négligeable (communication réseau par exemple).
# 
# La parallélisation est assez récente. Elle est née avec les premiers processeurs [multitâche](https://fr.wikipedia.org/wiki/Multit%C3%A2che) et les premiers [systèmes d'exploitation](https://fr.wikipedia.org/wiki/Syst%C3%A8me_d%27exploitation) avec des [interfaces graphiques](https://fr.wikipedia.org/wiki/Interface_graphique).
# 
# * [thread](https://fr.wikipedia.org/wiki/Thread_%28informatique%29) : [async, await](https://docs.python.org/3/whatsnew/3.5.html?highlight=await#whatsnew-pep-492), [queue](https://docs.python.org/3/library/queue.html?highlight=queue#module-queue), [threading](https://docs.python.org/3/library/threading.html?highlight=threading#module-threading)
# * [processus](https://fr.wikipedia.org/wiki/Processus_%28informatique%29) : [concurrent.futures](https://docs.python.org/3/library/concurrent.futures.html#module-concurrent.futures), [multiprocessing](https://docs.python.org/3/library/multiprocessing.html?highlight=pool#module-multiprocessing), [joblib](https://pythonhosted.org/joblib/)
# * [cluster](https://fr.wikipedia.org/wiki/Cluster)
# 

# ## joblib

# In[2]:


from math import sqrt
from joblib import Parallel, delayed
get_ipython().run_line_magic('timeit', 'Parallel(n_jobs=2)(delayed(sqrt)(i ** 2) for i in range(100000))')


# In[3]:


get_ipython().run_line_magic('timeit', '[sqrt(i ** 2) for i in range(100000)]')


# In[4]: